#
# Copyright (c) 2011-2013 LAAS/CNRS
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software for any purpose
# with or without   fee is hereby granted, provided   that the above  copyright
# notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS  SOFTWARE INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR  BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR  ANY DAMAGES WHATSOEVER RESULTING  FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION,   ARISING OUT OF OR IN    CONNECTION WITH THE USE   OR
# PERFORMANCE OF THIS SOFTWARE.
#
#                                            Anthony Mallet on Sun Jul 24 2011
#


# --- session2db -----------------------------------------------------------
#
proc session2db { dir db path } {
  infox "Collecting session logs in $dir"

  set count 0
  set entries [glob -no -dir [file normalize $dir] -types d *]

  set db [opendb $db]

  # iterate over log dirs
  foreach ent $entries {
    incr count [dir2db $ent $db $path]

    # next
    file delete -force $ent
  }

  infox "processed $count packages"
  if {[llength $entries] > $count} {
    warnx "ignored [expr {[llength $entries] - $count}] packages"
  }

  # log session, delete (old?) session tags with no pkg
  foreach log [glob -no -dir [file normalize $dir] -types f *.log] {
    set f [open $log]; set data [read $f]; close $f
    set tag [file rootname [file tail $log]]
    set date [file mtime $log]
    $db eval {
      insert into session (tag, log, date, updates, running)
      values ($tag, $data, $date, $count, 0)
    }
    file delete $log
  }
  $db eval { delete from session where tag not in (select tag from pkg) }

  closedb $db
  return
}


# --- log2db ---------------------------------------------------------------
#
proc log2db { dir db path } {
  if {![file isfile $dir]} {
    infox "Collecting package logs in $dir"
  }

  set db [opendb $db]
  if {![file isfile $dir]} {
    dir2db $dir $db $path
  }

  # log session progress
  if {![file isfile $dir]} {
    set log [glob -no -dir [file normalize $dir/..] -types f *.log]
  } else {
    set log $dir
  }
  if {[llength $log] != 1} {
    set progress -1
  } else {
    set f [open $log]; set data [read $f]; close $f
    set progress [string last "%] Processing" $data]
    if {$progress >= 3} {
      set data [string range $data $progress-3 $progress-1]
      set progress [regsub -all {[^0-9]} $data {}]
      if {$progress < 1} { set progress 1 }
    }
  }

  set count [glob -no -dir [file normalize $dir/..] -types f */meta.txt]
  set count [llength $count]

  # log current session state
  foreach log [glob -no -dir [file normalize $dir/..] -types f *.log] {
    set f [open $log]; set data [read $f]; close $f
    set tag [file rootname [file tail $log]]
    set date [file mtime $log]
    $db eval {
      insert into session (tag, log, date, updates, running)
      values ($tag, $data, $date, $count, $progress)
    }
  }

  closedb $db
  return
}


# --- dir2db ---------------------------------------------------------------
#
proc dir2db { dir db path } {
  cd $dir
  if {![file exists meta.txt]} {
    warnx "skipping non-bulk log directory \"[file tail $dir]\""
    return 0
  }

  # read meta data
  set f [open meta.txt]
  while {[gets $f v] >= 0} {
    foreach i [lrange $v 1 end] {
      lappend meta([lindex $v 0]) $i
    }
  }
  close $f
  if {[file exists cbbh.log]} { set meta(na) $meta(DATE_START) }
  if {[file exists broken.log]} { set meta(broken) $meta(DATE_START) }
  if {[file exists plist.log] || [file exists mirror.log]} {
    set meta(warning) $meta(DATE_START)
  }

  # skip unwanted paths
  if {![string match ${path}* $meta(PKGPATH)]} {
    infox "skipped $meta(PKGNAME)"
    return 0
  }

  # delete obsolete records
  if {[info exists meta(AVAIL)]} {
    $db eval [subst {
      select id from pkg natural join meta where
      tag = '$meta(BULK_TAG)' and path = '$meta(PKGPATH)'
      and pkgname not in ('[join $meta(AVAIL) {','}]')
    }] {
      $db eval { delete from pkg where id is $id }
      $db eval { delete from meta where id is $id }
      $db eval { delete from breaks where id is $id or breaks is $id}
      $db eval { delete from category where id is $id}
      $db eval { delete from log where id is $id }
    }
  }

  # record pkg
  $db eval {
    insert into pkg
    (tag, pkgname) values ($meta(BULK_TAG), $meta(PKGNAME))
  }
  set meta(id) [$db eval {
    select id from pkg where
    tag is $meta(BULK_TAG) and pkgname is $meta(PKGNAME)
  }]
  if {![info exists meta(na)]} {
    if {![info exists meta(broken)] && ![info exists meta(warning)]} {
      $db eval {select broken,warning from meta where id is $meta(id)} {
        if {$broken != 0 || $warning != 0} {
          set meta(fixed) $meta(DATE_START)
        }
      }
    }
  }

  # record brokenby
  $db eval { delete from breaks where breaks is $meta(id) }
  if {[file exists brokenby.log]} {
    set f [open brokenby.log]
    while {[gets $f v] >= 0} {
      $db eval [subst {
        select id from pkg where
        tag is '$meta(BULK_TAG)' and pkgname is '$v'
      }] {
        $db eval {
          insert into breaks (id, breaks) values ($id, $meta(id))
        }
      }
    }
    close $f
  }

  # record meta
  $db eval {
    insert into meta
    (id, path, pkgbase, na, date_start, date_stop, warning, broken, fixed)
    values ($meta(id), $meta(PKGPATH), $meta(PKGBASE), $meta(na),
            $meta(DATE_START), $meta(DATE_STOP), $meta(warning),
            $meta(broken), $meta(fixed))
  }
  $db eval { delete from category where id is $meta(id) }
  foreach c $meta(CATEGORIES) {
    $db eval {
      insert into category (id, category) values ($meta(id), $c)
    }
  }

  # record logs
  $db eval {delete from log where id is $meta(id)}
  if {[info exists meta(na)] ||
      [info exists meta(broken)] ||
      [info exists meta(warning)]} {
    foreach log [glob -no -types f *.log] {
      set f [open $log]; set data [read $f]; close $f
      $db eval {insert into log
        (id, file, data) values ($meta(id), $log, $data)
      }
    }
  }

  infox "updated $meta(PKGNAME)"
  return 1
}
