/*
 * Copyright (c) 2013 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Sat Jan 12 2013
 */
#include "acrbulkit.h"

#include <sys/param.h>
#include <sys/resource.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/wait.h>

#include <arpa/inet.h>
#include <netinet/in.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <fnmatch.h>
#include <grp.h>
#include <libgen.h>
#include <netdb.h>
#include <pwd.h>
#include <signal.h>
#include <spawn.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>


/* --- gobal data ---------------------------------------------------------- */

char config[MAXPATHLEN] = PACKAGE_PREFIX "/etc/rbulk-watchd.conf";
char *run_user = NULL;
char *run_group = NULL;
int daemonize = 1;
int daemonized = 0;
const char *port = "8869";
int dotcp = 1;
int doudp = 1;

char **bind_addr;
char **allow_addr;
char **trigger;

sigset_t dfltmask;
int max_socket;
fd_set socket_set;
fd_set socket_udp;
fd_set socket_tcp;
struct socket_state {
  pid_t pid;
  char ip[NI_MAXHOST], fqdn[NI_MAXHOST];
} **socket_state;

int leave = 0;


/* --- usage --------------------------------------------------------------- */

/** Print usage on a channel.
 *
 * \param[in] channel   Output channel
 * \param[in] argv0     Program invocation name
 */

static void
usage(FILE *channel, char *argv0)
{
  fprintf(
    channel,
    "usage:  %s [-h] [-t|-u] [-f] [-c file] [-p port] [-l user] [-g group]\n",
    basename(argv0));
}


/* --- log ----------------------------------------------------------------- */

static void
xlog(int prio, const char *fmt, ...)
{
  static int open = 0;
  va_list va;

  if (daemonize) {
    if (!open) {
      openlog("rbulk-watchd", LOG_CONS|LOG_PID, LOG_USER);
      open = 1;
    }
    va_start(va, fmt);
    vsyslog(prio, fmt, va);
    va_end(va);
  }
  if (!daemonize || (!daemonized && prio < LOG_INFO)) {
    va_start(va, fmt);
    vwarnx(fmt, va);
    va_end(va);
  }

  if (prio == LOG_ERR) {
    xlog(LOG_NOTICE, "terminating on fatal error");
    if (open) closelog();
    exit(2);
  }
}


/* --- noop_sighandler ----------------------------------------------------- */

static void
noop_sighandler(int sig)
{
}

static void
term_sighandler(int sig)
{
  leave = sig;
}


/* --- push_array ---------------------------------------------------------- */

static void
push_array(char ***array, const char *value)
{
  size_t n = (size_t)(*array)[0];

  n++;
  *array = realloc(*array, (n+1) * sizeof(**array));
  if (!*array) xlog(LOG_ERR, "malloc: %s", strerror(errno));

  (*array)[n] = strdup(value);
  (*array)[0] = (void *)n;
}


/* --- expand_state -------------------------------------------------------- */

static void
expand_state(int s, int create)
{
  int i;
  if (max_socket < s) {
    socket_state = realloc(socket_state, (s+1)*sizeof(*socket_state));
    if (!socket_state) xlog(LOG_ERR, "malloc: %s", strerror(errno));

    for(i = s; i > max_socket; i--)
      socket_state[i] = NULL;
    max_socket = s;
  }

  if (socket_state[s]) xlog(LOG_ERR, "trying to overwrite existing state");
  if (!create) return;

  socket_state[s] = calloc(1, sizeof(**socket_state));
  if (!socket_state[s]) xlog(LOG_ERR, "malloc: %s", strerror(errno));
}


/* --- load_config --------------------------------------------------------- */

#define RBW_WRD	"255[a-zA-Z0-9_/:-]"
#define RBW_PAT	"255[][*?a-zA-Z0-9_/:-]"

static int
load_config()
{
  FILE *f;
  char *line, key[256], value[256], *p;
  int rest;
  int len;
  int status = 0;
  int lineno = 0;

  f = fopen(config, "r");
  if (!f) xlog(LOG_ERR, "%s: %s", config, strerror(errno));

  len = 256;
  p = line = malloc(len);
  if (!line) xlog(LOG_ERR, "malloc: %s", strerror(errno));

  while (fgets(p, len - (p - line), f) != NULL) {
    ++lineno;
    p = strrchr(p, '\n');
    if (p == NULL) {
      if (strlen(line) >= len-1) {
        len *= 2;
        line = realloc(line, len);
        if (!line) xlog(LOG_ERR, "malloc: %s", strerror(errno));
        p = line + strlen(line);
        continue;
      }

      xlog(LOG_WARNING, "%s:%d: no final newline", config, lineno);
      status = 2;
    } else
      *p = '\0';
    if (p > line && p[-1] == '\\') { p--; continue; }
    if ((p = strchr(line, '#'))) *p = '\0';
    for (p = line; isspace((int)*p); p++);
    if (*p == '\0') { p = line; continue; }

    if (sscanf(p, "%" RBW_WRD " %255s %n", key, value, &rest) != 2) {
      xlog(LOG_WARNING, "%s:%d: syntax error", config, lineno);
      status = 2;
      p = line;
      continue;
    }

    if (!strcmp(key, "bind")) {
      push_array(&bind_addr, value);
    } else if (!strcmp(key, "allow")) {
      push_array(&allow_addr, value);
    } else if (!strcmp(key, "trigger")) {
      char t[256];
      if (sscanf(value, "%" RBW_PAT, t) != 1 || strlen(t) != strlen(value)) {
        xlog(LOG_WARNING, "%s:%d: invalid characters in `%s'",
             config, lineno, value);
        status = 2;
        p = line;
        continue;
      }
      if (line[rest] == '\0') {
        xlog(LOG_WARNING, "%s:%d: missing trigger action for %s",
             config, lineno, value);
        status = 2;
        p = line;
        continue;
      }
      push_array(&trigger, value);

      p = trigger[(size_t)trigger[0]];
      p = realloc(p, strlen(p) + 1 + sizeof(value));
      if (!p) xlog(LOG_ERR, "malloc: %s", strerror(errno));
      trigger[(size_t)trigger[0]] = p;
      p += strlen(value) + 1;
      strncpy(p, line + rest, sizeof(value));
      p[sizeof(value)-1] = '\0';
    } else {
      xlog(LOG_WARNING, "%s:%d: unknown key %s", config, lineno, key);
      status = 2;
      p = line;
      continue;
    }

    p = line;
    xlog(LOG_INFO, "configuration: %s %s%s%s",
         key, value, line[rest]?" ":"", line + rest);
  }

  free(line);
  fclose(f);
  return status;
}


/* --- trigger_command ----------------------------------------------------- */

static void
trigger_command(int s, const char *key)
{
  char *argp[] = { "sh", "-c", NULL, NULL };
  const char *p, *q;
  char *cmd, *c;
  pid_t pid;
  int len;
  int i;

  /* lookup first matching trigger */
  for(i=0; i<(size_t)trigger[0]; i++) {
    if (!fnmatch((trigger+1)[i], key, 0)) {
      cmd = (trigger+1)[i];
      cmd += strlen(cmd) + 1;
      break;
    }
  }
  if (i>=(size_t)trigger[0]) {
    xlog(LOG_WARNING, "no trigger for %s", key);
    if (!FD_ISSET(s, &socket_udp)) {
      close(s);
      free(socket_state[s]);
      socket_state[s] = NULL;
    }
    return;
  }

  /* build command - replace %s & %p */
  argp[2] = malloc(len = 64);
  for(c = argp[2], p = cmd; *p; p++) {
    if (c - argp[2] >= len-1) {
      argp[2] = realloc(argp[2], len*=2);
      if (!argp[2]) xlog(LOG_ERR, "malloc: %s", strerror(errno));
      c = argp[2] + len/2-1;
    }
    if (*p != '%') { *c++ = *p; continue; }
    p++;
    switch(*p) {
      case '\0': p--; q = ""; break;
      case '%': q = "%"; break;
      case 's': q = key; break;
      default: q = ""; break;
    }
    for(; *q; *c++ = *q++) {
      if (c - argp[2] >= len-1) {
        argp[2] = realloc(argp[2], len*=2);
        if (!argp[2]) xlog(LOG_ERR, "malloc: %s", strerror(errno));
        c = argp[2] + len/2-1;
      }
    }
  }
  *c = *"";

  /* spawn */
  send(s, argp[2], strlen(argp[2]), 0);
  send(s, "\n", sizeof("\n"), 0); /* include final \0 */
  switch((pid = fork())) {
    case -1:
      xlog(LOG_WARNING, "fork: %s", strerror(errno));
      if (!FD_ISSET(s, &socket_udp)) {
        close(s);
        free(socket_state[s]);
        socket_state[i] = NULL;
      }
      break;

    case 0: /* child */
      setsid();
      sigprocmask(SIG_SETMASK, &dfltmask, NULL);
      if (!FD_ISSET(s, &socket_udp)) {
        dup2(s, STDOUT_FILENO);
        dup2(s, STDERR_FILENO);
        if (s > STDERR_FILENO) close(s);
      }
      execvp(argp[0], argp);
      xlog(LOG_WARNING, "%s: %s", argp[0], strerror(errno));
      _exit(127);
  }

  /* remember the pid for tcp */
  if (!FD_ISSET(s, &socket_udp))
    socket_state[s]->pid = pid;
  xlog(LOG_NOTICE, "pid %d: running '%s'", pid, argp[2]);
  free(argp[2]);
}



/* --- setup_socket --------------------------------------------------------- */

static void
setup_socket(const char *bind_addr, int proto)
{
  struct addrinfo hints;
  struct addrinfo *res, *res0;
  int opt;
  int s;
  int e;

  /* lookup bind addr */
  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = (proto == IPPROTO_TCP) ? SOCK_STREAM:SOCK_DGRAM;
  hints.ai_protocol = proto;
  hints.ai_flags = AI_PASSIVE;
  if (bind_addr) hints.ai_flags |= AI_CANONNAME;
  e = getaddrinfo(bind_addr, port, &hints, &res0);
  if (e) xlog(LOG_ERR, "%s: %s.%s",
              gai_strerror(e), bind_addr?bind_addr:"<any>", port);

  /* socket */
  for(res = res0; res; res = res->ai_next) {
    s = socket(
      res->ai_family, res->ai_socktype|SOCK_CLOEXEC, res->ai_protocol);
    if (s < 0) xlog(LOG_ERR, "socket: %s", strerror(errno));

    if (res->ai_family == AF_INET6) {
      opt = 1;
      e = setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY, &opt, sizeof(opt));
      if (e) xlog(LOG_ERR, "IP6_V6ONLY: %s", strerror(errno));
    }

    opt = 1;
    e = setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
    if (e) xlog(LOG_ERR, "SO_REUSEADDR: %s", strerror(errno));

    if (bind(s, res->ai_addr, res->ai_addrlen) < 0)
      xlog(LOG_ERR, "binding on %s.%s: %s",
           res0->ai_canonname?res0->ai_canonname:"<any>", port,
           strerror(errno));

    if (res->ai_protocol == IPPROTO_TCP) {
      if (listen(s, 2))
        xlog(LOG_ERR, "listening on %s.%s: %s",
             res0->ai_canonname?res0->ai_canonname:"<any>", port,
             strerror(errno));
    }

    FD_SET(s, &socket_set);
    FD_SET(s, res->ai_protocol == IPPROTO_TCP ? &socket_tcp:&socket_udp);
    expand_state(s, 0/*create*/);

    xlog(LOG_INFO, "listening on %s %s %s.%s",
         res->ai_family == AF_INET ? "inet":"inet6",
         res->ai_protocol == IPPROTO_TCP ? "tcp":"udp",
         res0->ai_canonname?res0->ai_canonname:"<any>", port);
  }

  freeaddrinfo(res0);
}


/* --- auth_conn ----------------------------------------------------------- */

static int
auth_conn(int s, const struct sockaddr *from, socklen_t flen)
{
  size_t n = (size_t)allow_addr[0];
  struct socket_state state;
  int i, e;

  /* numeric address */
  e = getnameinfo(from, flen, state.ip, NI_MAXHOST, NULL, 0, NI_NUMERICHOST);
  if (e) {
    xlog(LOG_WARNING, "cannot handle incoming message: %s", gai_strerror(e));
    return 2;
  }
  for(i=0; i<n; i++) {
    if (!fnmatch((allow_addr+1)[i], state.ip, 0)) {
      *state.fqdn = '\0';
      goto pass;
    }
  }

  /* FQDN */
  e = getnameinfo(from, flen, state.fqdn, NI_MAXHOST, NULL, 0, NI_NAMEREQD);
  if (e) {
    xlog(LOG_NOTICE, "%s: connection refused", state.ip);
    return 2;
  }
  for(i=0; i<n; i++) {
    if (!fnmatch((allow_addr+1)[i], state.fqdn, 0))
      goto pass;
  }

  xlog(LOG_NOTICE, "%s[%s]: connection refused", state.fqdn, state.ip);
  return 2;

pass:
  expand_state(s, 1/*create*/);
  *socket_state[s] = state;
  return 0;
}


/* --- recv_notify --------------------------------------------------------- */

static void
recv_notify(void)
{
  char msg[65536];
  union {
    struct sockaddr_storage buf;
    struct sockaddr from;
  } from;
  socklen_t flen;
  fd_set set;
  ssize_t l;
  int e, i, s;
  char t[256], *p;

  /* recv messages or signals */
  memcpy(&set, &socket_set, sizeof(set));
  e = pselect(max_socket + 1, &set, NULL, NULL, NULL, &dfltmask);
  switch(e) {
    case -1:
      if (errno != EINTR)
        xlog(LOG_ERR, "waiting for notifications: %s", strerror(errno));
      break;

    case 0:
      xlog(LOG_WARNING, "timeout while waiting for notifications");
      break;

    default:
      for (i=0; i<=max_socket; i++) {
        if (!FD_ISSET(i, &set)) continue;

        /* tcp connection request */
        if (FD_ISSET(i, &socket_tcp)) {
          flen = sizeof(from);
          s = accept(i, &from.from, &flen);
          if (s < 0) {
            xlog(LOG_WARNING, "accept: %s", strerror(errno));
            continue;
          }
          if (auth_conn(s, &from.from, flen)) {
            close(s);
            continue;
          }
          send(s, "", 1, 0); /* notify acceptance */
          FD_SET(s, &socket_set);
          fcntl(s, F_SETFD, FD_CLOEXEC);
          fcntl(s, F_SETFL, O_NONBLOCK);
          xlog(LOG_NOTICE, "%s[%s]: connection accepted",
               socket_state[s]->fqdn, socket_state[s]->ip);
          continue;
        }

        /* udp datagram / tcp message */
        errno = 0;
        flen = sizeof(from);
        l = recvfrom(i, msg, sizeof(msg) - 1, 0, &from.from, &flen);
        if (l <= 0) {
          if (errno == EAGAIN || errno == EWOULDBLOCK) continue;
          if (!FD_ISSET(i, &socket_udp)) {
            /* reset tcp state */
            xlog(LOG_WARNING, "%s[%s]: %s",
                 socket_state[i]->fqdn, socket_state[i]->ip,
                 l ? strerror(errno):"connection closed");
            close(i);
            free(socket_state[i]);
            socket_state[i] = NULL;
            FD_CLR(i, &socket_set);
          }
          continue;
        }
        if (FD_ISSET(i, &socket_udp)) {
          /* auth every udp message */
          if (auth_conn(i, &from.from, flen)) continue;
        }

        /* trim whitespace */
        msg[l] = '\0';
        for(p = msg + l -1; isspace((int)*p); *p-- = '\0');
        for(p = msg; isspace((int)*p); p++);

        /* check input chars */
        if (sscanf(p, "%" RBW_WRD, t) != 1 || strlen(t) != strlen(p)) {
          xlog(LOG_WARNING, "%s[%s]: bad notification `%.255s'",
               socket_state[i]->fqdn, socket_state[i]->ip, t);
          continue;
        }

        if (!FD_ISSET(i, &socket_udp)) {
          /* tcp accepts only one message */
          shutdown(i, SHUT_RD);
          FD_CLR(i, &socket_set);
        }

        xlog(LOG_NOTICE, "%s[%s]: notified %.255s",
             socket_state[i]->fqdn, socket_state[i]->ip, p);
        trigger_command(i, p);

        if (FD_ISSET(i, &socket_udp)) {
          /* udp don't keep state */
          free(socket_state[i]);
          socket_state[i] = NULL;
        }
      }
      break;
  }
}


/* --- join_child ---------------------------------------------------------- */

static void
join_child(void)
{
  char msg[256];
  struct rusage rusage;
  pid_t p;
  int e, i;

  while ((p = wait4(-1, &e, WNOHANG, &rusage)) > 0) {
    /* fetch tcp state */
    for(i=0; i<=max_socket; i++)
      if (socket_state[i] && socket_state[i]->pid == p) break;
    if (i <= max_socket)
      send(i, "", 1, 0); /* signal end of process data */

    /* send the exit status */
    if (WIFSIGNALED(e)) {
      snprintf(msg, sizeof(msg), "%s%s",
               strsignal(WTERMSIG(e)), WCOREDUMP(e)?" (core dumped)":"");
      xlog(LOG_WARNING, "pid %d: %s", p, msg);
    } else if (WIFEXITED(e)) {
      snprintf(msg, sizeof(msg), "exit %d", WEXITSTATUS(e));
      if (WEXITSTATUS(e))
        xlog(LOG_WARNING, "pid %d: %s", p, msg);
      else
        xlog(LOG_INFO, "pid %d: exited normally", p);
    }
    if (i <= max_socket) {
      send(i, msg, strlen(msg), 0);
      send(i, "\n", sizeof("\n")-1, 0);
    }

    /* display ressource usage */
    snprintf(msg, sizeof(msg), "%gs user, %gs system",
             rusage.ru_utime.tv_sec+1e-6*rusage.ru_utime.tv_usec,
             rusage.ru_stime.tv_sec+1e-6*rusage.ru_stime.tv_usec);
    xlog(LOG_INFO, "pid %d: %s", p, msg);
    if (i <= max_socket) {
      send(i, msg, strlen(msg), 0);
      send(i, "\n", sizeof("\n")-1, 0);
    }

    /* cleanup state */
    if (i <= max_socket) {
      close(i);
      free(socket_state[i]);
      socket_state[i] = NULL;
    }
  }
}


/* --- main ---------------------------------------------------------------- */

int
main(int argc, char *const argv[])
{
  extern char *optarg;
  extern int optind;
  char *argv0 = argv[0];

  struct sigaction act;
  int c, n;

  /* parse command line options */
  while ((c = getopt(argc, argv, "htuc:p:fl:g:")) != -1) {
    switch (c) {
      case 'f':
        daemonize = 0;
        break;

      case 'c':
        if (access(optarg, R_OK) || !realpath(optarg, config))
          xlog(LOG_ERR, "%s: %s", optarg, strerror(errno));
        break;

      case 't':
        if (!dotcp) xlog(LOG_ERR, "options -t and -u are mutually exclusive");
        doudp = 0;
        break;

      case 'u':
        if (!doudp) xlog(LOG_ERR, "options -t and -u are mutually exclusive");
        dotcp = 0;
        break;

      case 'p':
        port = optarg;
        break;

      case 'l':
        run_user = optarg;
        break;

      case 'g':
        run_group = optarg;
        break;

      case 'h':
        usage(stdout, argv0);
        exit(0);

      case '?':
      default:
        usage(stderr, argv0);
        exit(2);
    }
  }
  if (argc > optind) {
    usage(stderr, argv0);
    exit(2);
  }

  if (run_group) {
    struct group *grp;
    errno = 0;
    grp = getgrnam(run_group);
    if (!grp) {
      if (errno) xlog(LOG_ERR, "%s: %s", run_group, strerror(errno));
      xlog(LOG_ERR, "%s: unkown group", run_group);
    }
    if (setgid(grp->gr_gid))
      xlog(LOG_ERR, "setgid: %s", strerror(errno));
    if (setgroups(0, NULL))
      xlog(LOG_ERR, "setgroups: %s", strerror(errno));
  }

  if (run_user) {
    struct passwd *pw;
    errno = 0;
    pw = getpwnam(run_user);
    if (!pw) {
      if (errno) xlog(LOG_ERR, "%s: %s", run_user, strerror(errno));
      xlog(LOG_ERR, "%s: unkown id", run_user);
    }
    if (setuid(pw->pw_uid))
      xlog(LOG_ERR, "setuid: %s", strerror(errno));
  }

restart:
  /* load config */
  bind_addr = malloc(sizeof(*bind_addr));
  if (!bind_addr) xlog(LOG_ERR, "malloc: %s", strerror(errno));
  bind_addr[0] = (char *)0;

  allow_addr = malloc(sizeof(*allow_addr));
  if (!allow_addr) xlog(LOG_ERR, "malloc: %s", strerror(errno));
  allow_addr[0] = (char *)0;

  trigger = malloc(sizeof(*trigger));
  if (!trigger) xlog(LOG_ERR, "malloc: %s", strerror(errno));
  trigger[0] = (char *)0;

  if (load_config())
    xlog(LOG_ERR, "cannot parse configuration in %s", config);


  /* setup sockets */
  FD_ZERO(&socket_set);
  FD_ZERO(&socket_tcp);
  FD_ZERO(&socket_udp);
  socket_state = calloc(1, sizeof(*socket_state));
  max_socket = 0;

  if (bind_addr[0]) {
    for(n=0; n<(size_t)bind_addr[0]; n++) {
      if (dotcp) setup_socket((bind_addr+1)[n], IPPROTO_TCP);
      if (doudp) setup_socket((bind_addr+1)[n], IPPROTO_UDP);
    }
  } else {
    if (dotcp) setup_socket(NULL, IPPROTO_TCP);
    if (doudp) setup_socket(NULL, IPPROTO_UDP);
  }


  /* daemonize */
  if (daemonize && !daemonized) {
    pid_t p;
    switch((p = fork())) {
      case -1:
        xlog(LOG_ERR, "cannot fork: %s", strerror(errno));

      case 0: /* child */ {
        int fd;

        if (setsid() == -1) exit(-1);
        if (chdir("/")) /*noop*/;

        fd = open("/dev/null", O_RDWR, 0);
        if (fd != -1) {
          (void)dup2(fd, STDIN_FILENO);
          (void)dup2(fd, STDOUT_FILENO);
          (void)dup2(fd, STDERR_FILENO);
          if (fd > STDERR_FILENO)
            (void)close(fd);
        }
        break;
      }

      default: /* parent */
        xlog(LOG_INFO, "daemonizing to pid %d", p);
        _exit(0);
    }

    daemonized = 1;
  }

  /* cleanup handlers */
  act.sa_sigaction = NULL;
  act.sa_handler = term_sighandler;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
  if (sigaction(SIGHUP, &act, NULL))
    xlog(LOG_ERR, "sigaction: %s", strerror(errno));
  if (sigaction(SIGINT, &act, NULL))
    xlog(LOG_ERR, "sigaction: %s", strerror(errno));
  if (sigaction(SIGQUIT, &act, NULL))
    xlog(LOG_ERR, "sigaction: %s", strerror(errno));
  if (sigaction(SIGTERM, &act, NULL))
    xlog(LOG_ERR, "sigaction: %s", strerror(errno));

  /* children handlers */
  act.sa_sigaction = NULL;
  act.sa_handler = noop_sighandler;
  sigemptyset(&act.sa_mask);
  act.sa_flags = SA_NOCLDSTOP|SA_RESTART;
  if (sigaction(SIGCHLD, &act, NULL))
    xlog(LOG_ERR, "sigaction: %s", strerror(errno));
  if (sigaction(SIGPIPE, &act, NULL))
    xlog(LOG_ERR, "sigaction: %s", strerror(errno));

  /* process signals in recv_notify() only */
  sigemptyset(&act.sa_mask);
  sigaddset(&act.sa_mask, SIGHUP);
  sigaddset(&act.sa_mask, SIGINT);
  sigaddset(&act.sa_mask, SIGQUIT);
  sigaddset(&act.sa_mask, SIGTERM);
  sigaddset(&act.sa_mask, SIGPIPE);
  sigaddset(&act.sa_mask, SIGCHLD);
  if (sigprocmask(SIG_BLOCK, &act.sa_mask, &dfltmask))
    xlog(LOG_ERR, "sigprocmask: %s", strerror(errno));

  /* handle notifications */
  while(!leave) {
    recv_notify();
    join_child();
  }

  /* cleanup */
  for (n=0; n<=max_socket; n++) {
    if (FD_ISSET(n, &socket_set)) close(n);
    if (socket_state[n]) free(socket_state[n]);
  }
  free(socket_state);
  for(n=0; n<(size_t)bind_addr[0]; n++)
    free((bind_addr+1)[n]);
  free(bind_addr);
  for(n=0; n<(size_t)allow_addr[0]; n++)
    free((allow_addr+1)[n]);
  free(allow_addr);
  for(n=0; n<(size_t)trigger[0]; n++)
    free((trigger+1)[n]);
  free(trigger);

  if (leave == SIGHUP) {
    xlog(LOG_NOTICE, "restarting on signal %d", leave);
    (void)sigprocmask(SIG_SETMASK, &dfltmask, NULL);
    leave = 0;
    goto restart;
  }

  xlog(LOG_NOTICE, "terminating on signal %d", leave);
  return 0;
}
